import 'dart:async';
import 'dart:convert';
import 'dart:developer';
import 'dart:io' show Platform;
import 'dart:io';

import 'package:collection/collection.dart';
import 'package:events2/events2.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_foreground_task/flutter_foreground_task.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:http/http.dart' as http;

import '../../../videosdk.dart';
import '../../core/pubsub/pubsub.dart';
import '../../core/webrtc/src/handlers/handler_interface.dart';
import '../../core/webrtc/webrtc_client.dart';
import '../../services/signalling/web_socket.dart';
import '../../utils/constants.dart';
import '../webrtc/src/utils.dart';
import 'events.dart';
import 'peer.dart';
import 'sdk_capabilites.dart';

class Room {
  // Room Configuration Properties
  late final String id;
  late final String _token;
  late final bool _micEnabled;
  late final bool _camEnabled;
  late final NotificationInfo _notification;
  late String _maxResolution;

  // Event Emitters
  final _eventEmitter = EventEmitter();
  final _topicEventEmitter = EventEmitter();

  // Participant
  late final Participant localParticipant;
  final participants = <String, Participant>{};

  // PubSub
  late PubSub pubSub;

  // States for MediaDevices
  MediaDeviceInfo? _selectedAudioInput;
  MediaDeviceInfo? _selectedVideoInput;

  // States
  bool _cameraInProgress = false;
  bool _screenShareInProgress = false;

  // audio consumption interval
  var _audioConsumption = 0;

  // Internal Components
  WebSocket? _webSocket;
  Device? _device;
  Transport? _sendTransport, _recvTransport;
  bool _produce = false;
  final _consume = true;
  bool _joined = false, _closed = false;
  String? _activeSpeakerId, _activePresenterId;
  Producer? _micProducer, _cameraProducer, _screenshareProducer;
  RTCVideoRenderer? _cameraRenderer, _screenshareRenderer;
  final List<RTCIceServer> _iceServers = [];
  final Map<String, Peer> _peers = {};

  //
  final String _initConfigUrl =
      "https://api.videosdk.live/infra/v1/meetings/sdk-init-config";

  // Usage
  final _usageIntervalDuration = 1;
  var _usageDataPointsPeerMaxRes = {};
  Timer? _usageInterval;

  //
  Room({
    required String meetingId,
    required String token,
    required String participantId,
    required String displayName,
    required bool micEnabled,
    required bool camEnabled,
    required String maxResolution,
    required NotificationInfo notification,
    int defaultCameraIndex = 0,
  }) {
    //
    id = meetingId;
    //
    _token = token;
    //
    _micEnabled = micEnabled;
    //
    _camEnabled = camEnabled;
    //
    _maxResolution = maxResolution;
    //
    _notification = notification;

    //
    localParticipant = Participant(
      id: participantId,
      displayName: displayName,
      isLocal: true,
      eventEmitter: _eventEmitter,
      enablePeerMic: _enablePeerMic,
      disablePeerMic: _disablePeerMic,
      enablePeerCamera: _enablePeerCamera,
      disablePeerCamera: _disablePeerCamera,
      removePeer: _removePeer,
      setConsumerQuality: setConsumerQuality,
      setViewPort: setViewPort,
    );

    //
    pubSub = PubSub(
      pubsubPublish: _pubsubPublish,
      pubsubSubscribe: _pubsubSubscribe,
      pubsubUnsubscribe: _pubsubUnsubscribe,
      topicEventEmitter: _topicEventEmitter,
    );

    //
    _selectedAudioInput =
        VideoSDK.mediaDevices[MediaDeviceType.audioInput]?.first;
    _selectedVideoInput =
        VideoSDK.mediaDevices[MediaDeviceType.videoInput]?[defaultCameraIndex];

    //
    _eventEmitter.on("peers-bloc-participant-joined", (Peer newPeer) {
      final newParticipant = Participant(
        id: newPeer.id,
        displayName: newPeer.displayName,
        isLocal: false,
        eventEmitter: _eventEmitter,
        enablePeerMic: _enablePeerMic,
        disablePeerMic: _disablePeerMic,
        enablePeerCamera: _enablePeerCamera,
        disablePeerCamera: _disablePeerCamera,
        removePeer: _removePeer,
        setConsumerQuality: setConsumerQuality,
        setViewPort: setViewPort,
      );
      participants[newPeer.id] = newParticipant;
      _eventEmitter.emit('participant-joined', newParticipant);
    });

    //
    _eventEmitter.on("peers-bloc-participant-left", (peerId) {
      participants.remove(peerId);
      _eventEmitter.emit('participant-left', peerId);
    });

    //
    _eventEmitter.on("pubsub-message", _handlePubSubMessage);

    //
    _eventEmitter.on("peers-bloc-presenter-changed", (peerId) {
      _activePresenterId = peerId;
      _eventEmitter.emit("presenter-changed", peerId);
    });
  }

  //
  String? get activeSpeakerId => _activeSpeakerId;

  //
  String? get activePresenterId => _activePresenterId;

  //
  Future<void> muteMic() => _disableMic();

  //
  Future<void> unmuteMic() => _enableMic();

  //
  Future<void> _disableMic() async {
    //
    if (_micProducer != null) {
      //
      String micId = _micProducer!.id;

      //
      _removeProducer(_micProducer!, _ProducerType.micProducer);
      _micProducer = null;

      try {
        //
        await _webSocket!.socket.request(
          'closeProducer',
          {
            'producerId': micId,
          },
        );
      } catch (error) {
        log("Error while closing remote audio producer $error");
      }
    } else {
      log("mic is not enabled");
    }
  }

  //
  Future<void> _enableMic() async {
    //
    if (_device?.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) == false) {
      return;
    }

    //
    MediaStream? audioStream;
    MediaStreamTrack? track;

    //
    try {
      //
      audioStream = await _createAudioStream();
      //
      track = audioStream.getAudioTracks().first;

      //
      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(opusStereo: 1, opusDtx: 1),
        stream: audioStream,
        appData: {
          'source': 'mic',
        },
        source: 'mic',
      );
    } catch (error) {
      //
      if (audioStream != null) {
        //
        await audioStream.dispose();
      }
    }
  }

  //
  Future<void> enableCam([bool dontCheckProgress = false]) async {
    if (_cameraInProgress && !dontCheckProgress) {
      return;
    }

    _cameraInProgress = true;

    if (_device?.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) == false) {
      return;
    }

    MediaStream? videoStream;
    MediaStreamTrack? track;

    try {
      // NOTE: prefer using h264
      RtpCodecCapability? codec = _device!.rtpCapabilities.codecs.firstWhere(
          (RtpCodecCapability c) {
        return c.mimeType.toLowerCase() == 'video/vp9' ||
            c.mimeType.toLowerCase() == 'video/vp8';
      },
          // (RtpCodecCapability c) => c.mimeType.toLowerCase() == 'video/h264',
          orElse: () =>
              throw 'desired vp9 codec+configuration is not supported');

      videoStream = await _createVideoStream();

      track = videoStream.getVideoTracks().first;

      _cameraInProgress = true;

      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(
          videoGoogleStartBitrate: 1000,
        ),
        encodings: kIsWeb
            ? [
                RtpEncodingParameters(
                    scalabilityMode: 'S3T3_KEY', scaleResolutionDownBy: 1.0),
              ]
            : [
                RtpEncodingParameters(scaleResolutionDownBy: 4.0),
                RtpEncodingParameters(scaleResolutionDownBy: 2.0),
                RtpEncodingParameters(scaleResolutionDownBy: 1.0),
              ],
        stream: videoStream,
        appData: {
          'source': 'webcam',
          'width': 1280,
          'height': 720,
          'encodings': [
            {'scaleResolutionDownBy': 4.0},
            {'scaleResolutionDownBy': 2.0},
            {'scaleResolutionDownBy': 1.0}
          ]
        },
        source: 'webcam',
        codec: codec,
      );
    } catch (error) {
      _cameraInProgress = false;

      if (videoStream != null) {
        await videoStream.dispose();
      }
    }
  }

  //
  Future<void> disableCam() async {
    //
    if (_cameraProducer != null) {
      //
      String? cameraId = _cameraProducer?.id;

      //
      _cameraInProgress = true;

      //
      _removeProducer(_cameraProducer!, _ProducerType.cameraProducer);
      _cameraProducer = null;

      try {
        //
        await _webSocket!.socket.request('closeProducer', {
          'producerId': cameraId,
        });
      } catch (error) {
        //
        log("Error while closing remote video producer $error");
      } finally {
        //
        _cameraInProgress = false;
      }
    } else {
      log("camera is not enabled!");
    }
  }

  //
  Future<void> enableScreenShare() async {
    //
    if (Platform.isAndroid) {
      //
      if (_screenShareInProgress) {
        //
        return;
      }

      try {
        //
        await FlutterForegroundTask.startService(
          notificationTitle: _notification.title,
          notificationText: _notification.message,
        );
      } catch (error) {
        //
        log("error => $error");
      }

      //
      _screenShareInProgress = true;

      //
      MediaStream? shareStream;
      MediaStreamTrack? track;

      try {
        //
        RtpCodecCapability? codec = _device!.rtpCapabilities.codecs.firstWhere(
            (RtpCodecCapability c) {
          return c.mimeType.toLowerCase() == 'video/vp9' ||
              c.mimeType.toLowerCase() == 'video/vp8';
        },
            // (RtpCodecCapability c) => c.mimeType.toLowerCase() == 'video/h264',
            orElse: () =>
                throw 'desired vp9 codec+configuration is not supported');

        //
        shareStream = await _createShareStream();

        //
        track = shareStream.getVideoTracks().first;

        //
        _screenShareInProgress = true;

        //
        _sendTransport!.produce(
          track: track,
          codecOptions: ProducerCodecOptions(
            videoGoogleStartBitrate: 1000,
          ),
          encodings: kIsWeb
              ? [
                  RtpEncodingParameters(
                      scalabilityMode: 'S3T3_KEY', scaleResolutionDownBy: 1.0),
                ]
              : [],
          stream: shareStream,
          appData: {'share': true},
          source: 'screen',
          codec: codec,
        );
      } catch (error) {
        //
        log('error enable share  error => $error');

        //
        _screenShareInProgress = false;

        //
        FlutterForegroundTask.stopService();

        //
        if (shareStream != null) {
          await shareStream.dispose();
        }
      }
    }
  }

  //
  Future<void> disableScreenShare() async {
    //
    if (Platform.isAndroid) {
      //
      if (_screenshareProducer != null) {
        //
        _screenShareInProgress = true;
        //
        String screenShareId = _screenshareProducer!.id;

        //
        _removeProducer(
            _screenshareProducer!, _ProducerType.screenshareProducer);

        //
        _screenshareProducer = null;

        try {
          //
          await _webSocket!.socket.request('closeProducer', {
            'producerId': screenShareId,
          });
        } catch (error) {
          //
          log("disableShare error => $error");
        } finally {
          //
          FlutterForegroundTask.stopService();

          //
          _screenShareInProgress = false;
        }
      } else {
        //
        log("Screenshare is not enabled !!");
      }
    }
  }

  //
  Future<void> join() async {
    if (_joined) {
      //
      log("Already Joined");
      return;
    }

    await VideoSDK.loadMediaDevices();

    //
    String baseUrl;
    try {
      //
      final jsonRes = await http.post(Uri.parse(_initConfigUrl), body: {
        'roomId': id,
      }, headers: {
        'Authorization': _token,
      });

      //
      if (jsonRes.statusCode != 200) {
        throw "Error while getting baseUrl";
      }

      //
      final jsonData = json.decode(jsonRes.body);
      final data = jsonData['data'];
      baseUrl = data['baseUrl'];

      //
      try {
        //
        String decryptedIceServer = decrypt(data['iceServers']);
        //
        List<dynamic> iceServers = json.decode(decryptedIceServer.substring(
            0, decryptedIceServer.lastIndexOf(']') + 1)) as List;
        //
        iceServers.asMap().forEach((key, value) {
          //
          if (value['url'].toString().contains("stun")) {
            //
            RTCIceServer iceServer = RTCIceServer(
                credentialType: RTCIceCredentialType.password,
                username: "",
                urls: [value['url']],
                credential: "");
            _iceServers.add(iceServer);
          } else {
            //
            RTCIceServer iceServer = RTCIceServer(
                credentialType: RTCIceCredentialType.password,
                username: value['username'],
                urls: [value['url']],
                credential: value['credential']);
            _iceServers.add(iceServer);
          }
        });
      } catch (error) {
        //
        log(error.toString());
      }
    } catch (err) {
      //
      log(err.toString());
      //
      throw "Error while getting baseUrl";
    }

    //
    _webSocket = WebSocket(
      baseUrl: baseUrl,
      peerId: localParticipant.id,
      meetingId: id,
      token: _token,
    );

    //
    _webSocket!.onOpen = _requestEntry;

    //
    _webSocket!.onFail = () {};

    //
    _webSocket!.onDisconnected = () {
      if (_sendTransport != null) {
        _sendTransport!.close();
        _sendTransport = null;
      }
      if (_recvTransport != null) {
        _recvTransport!.close();
        _recvTransport = null;
      }
    };

    //
    _webSocket!.onClose = () {
      //
      if (_closed) return;

      //
      _close();
    };

    //
    _webSocket!.onRequest = (request, accept, reject) async {
      switch (request['method']) {
        case "close":
          {
            _close();
            break;
          }

        case 'newConsumer':
          {
            if (!_consume) {
              reject(403, 'I do not want to consume');
              break;
            }
            try {
              _recvTransport!.consume(
                id: request['data']['id'],
                producerId: request['data']['producerId'],
                kind: RTCRtpMediaTypeExtension.fromString(
                    request['data']['kind']),
                rtpParameters:
                    RtpParameters.fromMap(request['data']['rtpParameters']),
                appData: Map<String, dynamic>.from(request['data']['appData']),
                peerId: request['data']['peerId'],
                accept: accept,
              );
            } catch (error) {
              rethrow;
            }
            break;
          }

        case 'enableMic':
          {
            try {
              _eventEmitter.emit('mic-requested', {
                // 'participantId': peerId,
                "accept": () {
                  _enableMic();
                },
                "reject": () {},
              });

              accept();
            } catch (error) {
              rethrow;
            }
            break;
          }
        case 'disableMic':
          {
            try {
              _disableMic();
              accept();
            } catch (error) {
              rethrow;
            }
            break;
          }
        case 'enableWebcam':
          {
            try {
              _eventEmitter.emit('webcam-requested', {
                // "participantId": peerId,
                "accept": () {
                  enableCam();
                },
                "reject": () {},
              });

              accept();
            } catch (error) {
              rethrow;
            }
            break;
          }
        case 'disableWebcam':
          {
            try {
              disableCam();
              accept();
            } catch (error) {
              rethrow;
            }
            break;
          }
        case 'dataUsage':
          {
            //final document to send
            var consumptionDoc = <String, dynamic>{};

            //? summation for this all consumers streams.
            num resolutionTotal = 0;

            _usageDataPointsPeerMaxRes.forEach((peerData, maxRes) {
              resolutionTotal += maxRes as num;

              //if maxRes not exists then set first
              if (consumptionDoc[maxRes.toString()] == null) {
                consumptionDoc[maxRes.toString()] = [];
              }

              consumptionDoc[maxRes.toString()].add(peerData);
            });

            //get rounded resolution
            resolutionTotal = _getRoundedResolution(resolutionTotal);

            var data = <String, dynamic>{
              "data": <String, dynamic>{
                "peerId": localParticipant.id,
                "consumption": consumptionDoc,
                "audioConsumption": _audioConsumption,
              },
              "totalResolution": resolutionTotal,
            };

            accept(data);

            //flush the data points after sending data
            _audioConsumption = 0;
            _usageDataPointsPeerMaxRes = {};
            break;
          }
        default:
          break;
      }
    };

    //
    _webSocket!.onNotification = (notification) async {
      switch (notification['method']) {
        case 'producerScore':
          {
            break;
          }

        case 'consumerClosed':
          {
            String consumerId = notification['data']['consumerId'];
            _removePeerConsumer(consumerId);

            break;
          }

        case 'consumerPaused':
          {
            // String consumerId = notification['data']['consumerId'];
            // peersBloc.add(PeerPausedConsumer(consumerId: consumerId));
            break;
          }

        case 'consumerResumed':
          {
            // String consumerId = notification['data']['consumerId'];
            // peersBloc.add(PeerResumedConsumer(consumerId: consumerId));
            break;
          }

        case "consumerLayersChanged":
          {
            String consumerId = notification['data']['consumerId'];
            int spatialLayer = notification['data']['spatialLayer'] ?? -1;
            int temporalLayer = notification['data']['temporalLayer'] ?? -1;

            _changePeerConsumerQuality(
              consumerId,
              spatialLayer: spatialLayer,
              temporalLayer: temporalLayer,
            );

            Consumer? consumer = _recvTransport?.getConsumer(consumerId);

            if (consumer != null) {
              consumer.spatialLayer = spatialLayer;
              consumer.temporalLayer = temporalLayer;
            }

            break;
          }

        case "entryRequested":
          {
            var id = notification['data']['id'];
            var name = notification['data']['name'];

            if (id != localParticipant.id) {
              _eventEmitter.emit("entry-requested", <String, dynamic>{
                'participantId': id,
                'name': name,
                'allow': () {
                  _respondEntry(id, "allowed");
                },
                'deny': () {
                  _respondEntry(id, "denied");
                },
              });
            }

            break;
          }

        case "entryResponded":
          {
            var id = notification['data']['id'];
            var decision = notification['data']['decision'];

            _eventEmitter.emit(
              "entry-responded",
              <String, dynamic>{
                "id": id,
                "decision": decision,
              },
            );

            if (id == localParticipant.id && decision == "allowed") {
              _joinRoom();
            }

            break;
          }

        case 'newPeer':
          {
            final Map<String, dynamic> newPeer =
                Map<String, dynamic>.from(notification['data']);
            _addPeer(newPeer);
            break;
          }

        case 'peerClosed':
          {
            String peerId = notification['data']['peerId'];
            __removePeer(peerId);
            break;
          }

        case 'recordingStarted':
          {
            _eventEmitter.emit('recording-started');
            break;
          }

        case 'recordingStopped':
          {
            _eventEmitter.emit('recording-stopped');
            break;
          }

        case "livestreamStarted":
          {
            _eventEmitter.emit('livestream-started');

            break;
          }

        case "livestreamStopped":
          {
            _eventEmitter.emit('livestream-stopped');

            break;
          }

        case "activeSpeaker":
          {
            String? peerId = notification['data']['peerId'];

            if (peerId != _activeSpeakerId) {
              _activeSpeakerId = peerId;
              _eventEmitter.emit("speaker-changed", peerId);
            }

            break;
          }
        case "pubsubMessage":
          {
            _eventEmitter.emit("pubsub-message", notification['data']);
            break;
          }

        default:
          break;
      }
    };
  }

//
  Future<void> _joinRoom() async {
    try {
      _device = Device();

      dynamic routerRtpCapabilities =
          await _webSocket!.socket.request('getRouterRtpCapabilities', {});

      final rtpCapabilities = RtpCapabilities.fromMap(routerRtpCapabilities);
      rtpCapabilities.headerExtensions
          .removeWhere((he) => he.uri == 'urn:3gpp:video-orientation');
      await _device!.load(routerRtpCapabilities: rtpCapabilities);

      if (_device!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) == true ||
          _device!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) == true) {
        _produce = true;
      }

      if (_produce) {
        Map transportInfo =
            await _webSocket!.socket.request('createWebRtcTransport', {
          'forceTcp': false,
          'producing': true,
          'consuming': false,
          'sctpCapabilities': _device!.sctpCapabilities.toMap(),
        });

        _sendTransport = _device!.createSendTransportFromMap(
          transportInfo,
          producerCallback: _producerCallback,
          iceServers: _iceServers,
        );

        _sendTransport!.on('connect', (Map data) {
          _webSocket!.socket
              .request('connectWebRtcTransport', {
                'transportId': _sendTransport!.id,
                'dtlsParameters': data['dtlsParameters'].toMap(),
              })
              .then(data['callback'])
              .catchError(data['errback']);
        });

        _sendTransport!.on('produce', (Map data) async {
          try {
            Map response = await _webSocket!.socket.request(
              'produce',
              {
                'transportId': _sendTransport!.id,
                'kind': data['kind'],
                'rtpParameters': data['rtpParameters'].toMap(),
                if (data['appData'] != null)
                  'appData': Map<String, dynamic>.from(data['appData'])
              },
            );

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });

        _sendTransport!.on('producedata', (data) async {
          try {
            Map response = await _webSocket!.socket.request('produceData', {
              'transportId': _sendTransport!.id,
              'sctpStreamParameters': data['sctpStreamParameters'].toMap(),
              'label': data['label'],
              'protocol': data['protocol'],
              'appData': data['appData'],
            });

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });
      }

      if (_consume) {
        Map transportInfo = await _webSocket!.socket.request(
          'createWebRtcTransport',
          {
            'forceTcp': false,
            'producing': false,
            'consuming': true,
            'sctpCapabilities': _device!.sctpCapabilities.toMap(),
          },
        );

        _recvTransport = _device!.createRecvTransportFromMap(
          transportInfo,
          consumerCallback: _consumerCallback,
          iceServers: _iceServers,
        );

        _recvTransport!.on(
          'connect',
          (data) {
            _webSocket!.socket
                .request(
                  'connectWebRtcTransport',
                  {
                    'transportId': _recvTransport!.id,
                    'dtlsParameters': data['dtlsParameters'].toMap(),
                  },
                )
                .then(data['callback'])
                .catchError(data['errback']);
          },
        );
      }

      final deviceInfo = await VideoSDK.getDeviceInfo();

      Map response = await _webSocket!.socket.request('join', {
        'displayName': localParticipant.displayName,
        'device': {
          'name': "Flutter",
          'flag': 'flutter',
          'version': sdkVersion,
        },
        'rtpCapabilities': _device!.rtpCapabilities.toMap(),
        'sctpCapabilities': _device!.sctpCapabilities.toMap(),
        'deviceInfo': deviceInfo,
      });

      response['peers'].forEach((value) {
        _addPeer(value);
      });

      if (_produce) {
        if (_micEnabled) {
          _enableMic();
        }
        if (_camEnabled) {
          enableCam();
        }

        _sendTransport!.on('connectionstatechange', (connectionState) {
          if (connectionState == 'connected') {
            // enableChatDataProducer();
            // enableBotDataProducer();
          }
        });
      }

      _usageInterval = Timer.periodic(
        Duration(seconds: _usageIntervalDuration),
        (timer) => _updateUsageStats(),
      );

      _eventEmitter.emit('meeting-joined');
      _joined = true;

      try {
        await FlutterForegroundTask.init(
          androidNotificationOptions: AndroidNotificationOptions(
            channelId: 'notification_share_screen',
            channelName: 'Foreground Notification',
            channelDescription:
                'This notification appears when the foreground service is running.',
            channelImportance: NotificationChannelImportance.LOW,
            priority: NotificationPriority.LOW,
            iconData: NotificationIconData(
              resType: ResourceType.drawable,
              resPrefix: ResourcePrefix.ic,
              name: _notification.icon,
            ),
          ),
          iosNotificationOptions: const IOSNotificationOptions(
            showNotification: true,
            playSound: false,
          ),
          foregroundTaskOptions: const ForegroundTaskOptions(
            interval: 5000,
            autoRunOnBoot: true,
            allowWifiLock: true,
          ),
          printDevLog: true,
        );
      } catch (err) {
        log(err.toString());
      }
    } catch (error) {
      log(error.toString());
      _close();
    }
  }

  void end() {
    if (_closed) {
      return;
    }
    _webSocket!.socket.request('closeRoom', {});
    _eventEmitter.emit("meeting-left");
    _closed = true;
  }

  void leave() {
    _close();
  }

  void _close() {
    if (_closed) {
      return;
    }

    _webSocket?.close();
    _sendTransport?.close();
    _recvTransport?.close();

    if (_usageInterval?.isActive ?? false) _usageInterval?.cancel();
    _eventEmitter.emit("meeting-left");
    _closed = true;
  }

  Future<void> _requestEntry() async {
    await _webSocket!.socket.request("requestEntry", <String, dynamic>{
      'name': localParticipant.displayName,
    });
  }

  Future<void> _respondEntry(peerId, decision) async {
    await _webSocket!.socket.request(
        "respondEntry", <String, dynamic>{'id': peerId, 'decision': decision});
  }

  Future<void> startRecording(String webhookUrl) async {
    try {
      await _webSocket!.socket.request('startRecording', {
        'webhookUrl': webhookUrl,
      });
    } catch (error) {
      log("Error while starting recording $error");
    }
  }

  Future<void> stopRecording() async {
    try {
      await _webSocket!.socket.request('stopRecording', {});
    } catch (error) {
      log("Error while stopping recording $error");
    }
  }

  Future<void> startLivestream(outputs) async {
    try {
      await _webSocket!.socket.request('startLivestream', {"outputs": outputs});
    } catch (error) {
      log("Error while starting livestreaming $error");
    }
  }

  Future<void> stopLivestream() async {
    try {
      await _webSocket!.socket.request('stopLivestream', {});
    } catch (error) {
      log("Error while stopping livestreaming $error");
    }
  }

  Future<void> changeCam(String deviceId) async {
    if (_selectedVideoInput!.deviceId == deviceId) {
      return;
    }

    try {
      //
      _cameraInProgress = true;
      //
      MediaDeviceInfo device =
          VideoSDK.mediaDevices[MediaDeviceType.videoInput]!.firstWhere(
        (device) => device.deviceId == deviceId,
      );

      //
      _selectedVideoInput = device;

      //
      await disableCam();

      //
      enableCam();

      //
      _cameraInProgress = false;
    } catch (err) {
      //
      log("handle camera error $err");
    }
  }

  //
  List<MediaDeviceInfo> getCameras() =>
      VideoSDK.mediaDevices[MediaDeviceType.videoInput]!;

  //
  String? get selectedCamId => _selectedVideoInput?.deviceId;

  //
  List<MediaDeviceInfo> getMics() =>
      VideoSDK.mediaDevices[MediaDeviceType.audioInput]!;

  //
  String? get selectedMicId => _selectedAudioInput!.deviceId;

  //
  bool get micEnabled => _micEnabled;

  //
  bool get camEnabled => _camEnabled;

  //
  Future<void> _enablePeerMic(String peerId) async {
    try {
      await _webSocket!.socket.request('enablePeerMic', {
        "peerId": peerId,
      });
    } catch (error) {
      log("Error while enabling peer's mic $error");
    }
  }

  //
  Future<void> _disablePeerMic(String peerId) async {
    try {
      await _webSocket!.socket.request('disablePeerMic', {
        "peerId": peerId,
      });
    } catch (error) {
      log("Error while disabling peer's mic $error");
    }
  }

  //
  Future<void> _enablePeerCamera(String peerId) async {
    try {
      await _webSocket!.socket.request('enablePeerWebcam', {
        "peerId": peerId,
      });
    } catch (error) {
      log("Error while enabling peer's camera $error");
    }
  }

  //
  Future<void> _disablePeerCamera(String peerId) async {
    try {
      await _webSocket!.socket.request('disablePeerWebcam', {
        "peerId": peerId,
      });
    } catch (error) {
      log("Error while disabling peer's camera $error");
    }
  }

  //
  Future<void> _removePeer(String peerId) async {
    try {
      await _webSocket!.socket.request('removePeer', {
        "peerId": peerId,
      });
    } catch (error) {
      log("Error while removing peer $error");
    }
  }

  Future<void> _addProducer(Producer producer) async {
    final Consumer consumer = Consumer(
      peerId: localParticipant.id,
      appData: producer.appData,
      localId: producer.localId,
      producerId: producer.id,
      id: producer.id,
      stream: producer.stream,
      rtpParameters: producer.rtpParameters,
      track: producer.track,
      closed: producer.closed,
      rtpReceiver: null,
    );

    switch (producer.source) {
      case 'mic':
        {
          _eventEmitter.emit(
            "stream-enabled-${localParticipant.id}",
            {
              "audio": consumer,
            },
          );

          _micProducer = producer;
          break;
        }
      case 'webcam':
        {
          _cameraRenderer = RTCVideoRenderer();
          await _cameraRenderer?.initialize();

          _cameraRenderer?.srcObject = consumer.stream;

          _eventEmitter.emit(
            "stream-enabled-${localParticipant.id}",
            {
              "video": consumer,
              "renderer": _cameraRenderer as RTCVideoRenderer,
            },
          );

          _cameraProducer = producer;
          break;
        }
      case 'screen':
        {
          _screenshareRenderer = RTCVideoRenderer();
          await _screenshareRenderer!.initialize();

          _screenshareRenderer!.srcObject = consumer.stream;

          _eventEmitter.emit(
            "stream-enabled-${localParticipant.id}",
            {
              "share": consumer,
              "shareRenderer": _screenshareRenderer as RTCVideoRenderer,
            },
          );

          _screenshareProducer = producer;
          break;
        }
    }
  }

  void _removeProducer(Producer producer, _ProducerType type) {
    String? producerId = producer.id;
    switch (type) {
      case _ProducerType.micProducer:
        producer.close();

        _eventEmitter.emit(
          "stream-disabled-${localParticipant.id}",
          {
            "audioConsumerId": producerId,
          },
        );
        break;
      case _ProducerType.cameraProducer:
        //
        producer.close();

        //
        _eventEmitter.emit(
          "stream-disabled-${localParticipant.id}",
          {
            "renderer": _cameraRenderer,
            "videoConsumerId": producerId,
          },
        );
        break;
      case _ProducerType.screenshareProducer:

        //
        producer.close();

        //
        _eventEmitter.emit(
          "stream-disabled-${localParticipant.id}",
          {
            "shareRenderer": _screenshareRenderer,
            "shareConsumerId": producerId,
          },
        );
        break;
    }
  }

  // Peer
  //
  void _addPeer(Map<String, dynamic> newPeer) {
    //
    final Peer peer = Peer.fromMap(newPeer);
    //
    _peers[peer.id] = peer;
    //
    _eventEmitter.emit("peers-bloc-participant-joined", peer);
  }

  //
  void __removePeer(String peerId) {
    //
    _peers.remove(peerId);

    //
    _eventEmitter.emit("peers-bloc-participant-left", peerId);
  }

  //
  Future<void> _addPeerConsumer(String peerId, Consumer consumer) async {
    if (consumer.kind == 'video' && consumer.appData['share'] == true) {
      //
      _peers[peerId] = _peers[peerId]!.copyWith(
        shareRenderer: RTCVideoRenderer(),
        share: consumer,
      );

      //
      await _peers[peerId]!.shareRenderer!.initialize();

      //
      _peers[peerId]!.shareRenderer!.srcObject = _peers[peerId]!.share!.stream;

      //
      _eventEmitter.emit("stream-enabled-$peerId", {
        "shareRenderer": _peers[peerId]!.shareRenderer,
        "share": _peers[peerId]!.share
      });

      //
      _eventEmitter.emit("peers-bloc-presenter-changed", peerId);
    } else if (consumer.kind == 'video') {
      //
      _peers[peerId] = _peers[peerId]!.copyWith(
        renderer: RTCVideoRenderer(),
        video: consumer,
      );

      //
      await _peers[peerId]!.renderer!.initialize();

      //
      _peers[peerId]!.renderer!.srcObject = _peers[peerId]!.video!.stream;

      //
      _eventEmitter.emit("stream-enabled-$peerId", {
        "renderer": _peers[peerId]!.renderer,
        "video": _peers[peerId]!.video
      });
      //
    } else if (consumer.kind == 'audio') {
      _peers[peerId] = _peers[peerId]!.copyWith(
        audio: consumer,
      );

      _eventEmitter.emit("stream-enabled-$peerId", {
        "audio": consumer,
      });
    }
  }

  //
  Future<void> _removePeerConsumer(String consumerId) async {
    //
    final Peer? peer = _peers.values.firstWhereOrNull(
      (p) => p.consumers.contains(
        consumerId,
      ),
    );

    //
    if (peer != null) {
      if (peer.audio?.id == consumerId) {
        final consumer = peer.audio;
        _peers[peer.id] = _peers[peer.id]!.removeAudio();

        await consumer?.close();

        _eventEmitter.emit(
          "stream-disabled-${peer.id}",
          {
            "audioConsumerId": consumer?.id,
          },
        );
      } else if (peer.video?.id == consumerId) {
        final consumer = peer.video;
        final renderer = peer.renderer;
        _peers[peer.id] = _peers[peer.id]!.removeVideoAndRenderer();

        await consumer?.close();

        await renderer?.dispose();

        _eventEmitter.emit(
          "stream-disabled-${peer.id}",
          {
            "renderer": renderer,
            "videoConsumerId": consumer?.id,
          },
        );
      } else if (peer.share?.id == consumerId) {
        final consumer = peer.share;
        final renderer = peer.shareRenderer;

        _peers[peer.id] = _peers[peer.id]!.removeShareAndRenderer();

        await consumer?.close();

        await renderer?.dispose();

        _eventEmitter.emit(
          "stream-disabled-${peer.id}",
          {
            "shareRenderer": renderer,
            "shareConsumerId": consumer?.id,
          },
        );

        _eventEmitter.emit("peers-bloc-presenter-changed");
      }
    }
  }

  //
  // ignore: unused_element
  Future<void> _pausePeerConsumer(String consumerId) async {
    //
    final Peer? peer = _peers.values.firstWhereOrNull(
      (p) => p.consumers.contains(
        consumerId,
      ),
    );

    //
    if (peer != null) {
      //
      if (peer.audio?.id == consumerId) {
        //
        Peer newPeer = _peers[peer.id]!.copyWith(
          audio: peer.audio!.pauseCopy(),
        );

        //
        _peers[peer.id] = newPeer;

        //
        _eventEmitter.emit("stream-paused-${peer.id}", {
          "audioConsumerId": consumerId,
          "consumer": newPeer.audio,
        });
      } else if (peer.video?.id == consumerId) {
        //
        Peer newPeer = _peers[peer.id]!.copyWith(
          video: peer.video!.pauseCopy(),
        );

        //
        _peers[peer.id] = newPeer;

        //
        _eventEmitter.emit("stream-paused-${peer.id}", {
          "videoConsumerId": consumerId,
          "consumer": newPeer.video,
        });
      } else if (peer.share?.id == consumerId) {
        //
        Peer newPeer = _peers[peer.id]!.copyWith(
          share: peer.share!.pauseCopy(),
        );

        //
        _peers[peer.id] = newPeer;

        //
        _eventEmitter.emit("stream-paused-${peer.id}", {
          "shareConsumerId": consumerId,
          "consumer": newPeer.share,
        });
      }
    }
  }

  //
  // ignore: unused_element
  Future<void> _resumePeerConsumer(String consumerId) async {
    final Peer? peer = _peers.values.firstWhereOrNull(
      (p) => p.consumers.contains(
        consumerId,
      ),
    );

    if (peer != null) {
      if (peer.audio?.id == consumerId) {
        Peer newPeer = _peers[peer.id]!.copyWith(
          audio: peer.audio!.resumeCopy(),
        );

        _peers[peer.id] = newPeer;

        _eventEmitter.emit("stream-resumed-${peer.id}", {
          "audioConsumerId": consumerId,
          "consumer": newPeer.audio,
        });
      } else if (peer.video?.id == consumerId) {
        Peer newPeer = _peers[peer.id]!.copyWith(
          video: peer.video!.resumeCopy(),
        );

        _peers[peer.id] = newPeer;

        _eventEmitter.emit("stream-resumed-${peer.id}", {
          "videoConsumerId": consumerId,
          "consumer": newPeer.video,
        });
      } else if (peer.share?.id == consumerId) {
        Peer newPeer = _peers[peer.id]!.copyWith(
          share: peer.share!.resumeCopy(),
        );

        _peers[peer.id] = newPeer;

        _eventEmitter.emit("stream-resumed-${peer.id}", {
          "shareConsumerId": consumerId,
          "consumer": newPeer.share,
        });
      }
    }
  }

  //
  void _changePeerConsumerQuality(
    String consumerId, {
    required int spatialLayer,
    required int temporalLayer,
  }) {
    //
    final Peer? peer = _peers.values.firstWhereOrNull(
      (p) => p.consumers.contains(
        consumerId,
      ),
    );

    //
    if (peer != null) {
      //
      if (peer.video?.id == consumerId) {
        //
        Peer newPeer = _peers[peer.id]!.copyWith(
            video: peer.video!.copyWith(
          spatialLayer: spatialLayer,
          temporalLayer: temporalLayer,
        ));

        //
        _peers[peer.id] = newPeer;

        //
        _eventEmitter.emit("quality-changed-${peer.id}", {
          "spatialLayer": spatialLayer,
          "temporalLayer": temporalLayer,
          "videoConsumerId": consumerId
        });
      }
    }
  }

  //
  void _producerCallback(Producer producer) {
    if (producer.source == 'mic') {
      producer.on('transportclose', () {
        _removeProducer(_micProducer!, _ProducerType.micProducer);
      });

      producer.on('trackended', () {
        _disableMic().catchError((data) {});
      });
    } else if (producer.source == 'webcam') {
      producer.on('transportclose', () {
        _removeProducer(_cameraProducer!, _ProducerType.cameraProducer);
      });

      producer.on('trackended', () {
        disableCam().catchError((data) {});
      });
      _cameraInProgress = false;
    } else if (producer.source == 'share') {
      producer.on('transportclose', () {
        FlutterForegroundTask.stopService();
        _removeProducer(
            _screenshareProducer!, _ProducerType.screenshareProducer);
      });

      producer.on('trackended', () {
        FlutterForegroundTask.stopService();
        disableScreenShare().catchError((data) {});
      });
      _screenShareInProgress = false;
    }

    _addProducer(producer);
  }

  void _consumerCallback(Consumer consumer, [dynamic accept]) {
    accept({});

    _addPeerConsumer(consumer.peerId, consumer);
  }

  Future<MediaStream> _createAudioStream() async {
    //
    Map<String, dynamic> mediaConstraints = {
      'audio': {
        'optional': [
          {
            'sourceId': _selectedAudioInput!.deviceId,
          },
        ],
      },
    };

    //
    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    //
    return stream;
  }

  Future<MediaStream> _createVideoStream() async {
    //
    Map<String, dynamic> mediaConstraints = <String, dynamic>{
      'audio': false,
      'video': {
        'mandatory': {
          'minWidth':
              '1280', // Provide your own width, height and frame rate here
          'minHeight': '720',
          'minFrameRate': '30',
        },
        'optional': [
          {
            'sourceId': _selectedVideoInput?.deviceId,
          },
        ],
      },
    };
    //
    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    //
    return stream;
  }

  Future<MediaStream> _createShareStream() async {
    //
    final mediaConstraints = <String, dynamic>{'audio': true, 'video': true};

    //
    MediaStream stream =
        await navigator.mediaDevices.getDisplayMedia(mediaConstraints);

    //
    return stream;
  }

  // PubSub
  Future<void> _pubsubPublish({topic, message, options}) async {
    try {
      await _webSocket!.socket.request('pubsubPublish', {
        'topic': topic,
        'message': message,
        'options': options,
      });
    } catch (error) {
      log("pubsubPublish() | failed: $error");
    }
  }

  Future<dynamic> _pubsubSubscribe(topic) async {
    try {
      var msgList =
          await _webSocket!.socket.request('pubsubSubscribe', {'topic': topic});
      return msgList;
    } catch (error) {
      log("pubsubSubscribe() | failed: $error");
    }
  }

  Future<void> _pubsubUnsubscribe(topic) async {
    try {
      await _webSocket!.socket.request('pubsubUnsubscribe', {'topic': topic});
    } catch (error) {
      log("pubsubUnsubscribe() | failed: $error");
    }
  }

  void _handlePubSubMessage(data) {
    _topicEventEmitter.emit(
      data['topic'],
      PubSubMessage.fromJson(data),
    );
  }

  Future<void> setConsumerQuality(consumerId, String quality) async {
    late int? spatialLayers;
    late int? temporalLayers;

    try {
      if (quality == 'low' || quality == 'med' || quality == 'high') {
        final layers = VIDEO_LAYERS[_maxResolution]![quality];
        spatialLayers = layers?['s'];
        temporalLayers = layers?['t'];
      } else {
        String sCh = quality[0];
        String tCh = quality[2];
        int s = int.parse(quality[1]);
        int t = int.parse(quality[3]);

        if (s > 2 || t > 2 || sCh != "s" || tCh != "t") {
          throw ("");
        }

        spatialLayers = s;
        temporalLayers = t;
      }

      _setConsumerPreferredLayers(consumerId, spatialLayers, temporalLayers);
    } catch (error) {
      log("invalid quality");
    }
  }

  Future<void> _setConsumerPreferredLayers(
    consumerId,
    spatialLayer,
    temporalLayer,
  ) async {
    try {
      await _webSocket!.socket.request('setConsumerPreferredLayers', {
        "consumerId": consumerId,
        "spatialLayer": spatialLayer,
        "temporalLayer": temporalLayer,
      });
    } catch (error) {
      log("Error while setting consumer preferred layers $error");
    }
  }

  void setViewPort(consumerId, viewportWidth, viewportHeight) {
    Consumer? consumer = _recvTransport?.getConsumer(consumerId);
    if (consumer == null) {
      return;
    }
    Map<String, dynamic>? layers = SdkCapabilities.getAdaptivePreferredLayers(
      consumer,
      viewportWidth,
      viewportHeight,
    );
    if (layers != null) {
      _setConsumerPreferredLayers(
        consumerId,
        layers['newPreferredSpatialLayer'],
        layers['newPreferredTemporalLayer'],
      );
    }
  }

  on(Events event, Function handler) {
    _eventEmitter.on(event.parseToString(), handler);
  }

// Update usage
  _updateUsageStats() {
    if (_recvTransport?.consumers.isEmpty ?? true) return;

    List<Map<String, dynamic>> onUpdateUsageMap = [];

    _recvTransport?.consumers.forEach((consumerId, consumer) {
      if (consumer.kind == "video" && !consumer.paused) {
        var isScreenShare = consumer.appData['share'] ?? false;
        // ignore: prefer_typing_uninitialized_variables
        var peerData;

        var selectedResolution =
            _getResolutionFromSpatialLayer(consumer.spatialLayer);

        num trackResolutionRounded = _getRoundedResolution(selectedResolution);

        // check if this is peer's video stream or screen share
        if (isScreenShare) {
          peerData = "${consumer.appData['peerId']}-screen";
        } else {
          peerData = consumer.appData['peerId'];
        }

        //check if it's not exists then set this resolution
        if (!_usageDataPointsPeerMaxRes.containsKey(peerData)) {
          _usageDataPointsPeerMaxRes[peerData] = trackResolutionRounded;
        } else {
          //if exists then check if it's greater than the current max

          if (_usageDataPointsPeerMaxRes[peerData] < trackResolutionRounded) {
            _usageDataPointsPeerMaxRes[peerData] = trackResolutionRounded;
          }
        }
        onUpdateUsageMap.add(<String, dynamic>{
          "peerId": peerData,
          "resolution": trackResolutionRounded,
        });
      } else if (consumer.kind == "audio" && !consumer.paused) {
        _audioConsumption++;
      }
    });
  }

  int _getResolutionFromSpatialLayer(int spatialLayer) {
    switch (spatialLayer) {
      case -1:
        return 0;
      case 0:
        return 180;
      case 1:
        return 360;
      case 2:
        return 720;
      case 3:
        return 1080;
      default:
        return 360;
    }
  }

  num _getRoundedResolution(num? resolution) {
    if ((resolution == null) || (resolution == 0)) {
      return 0;
    } else if (resolution <= 180) {
      return 180;
    } else if (resolution > 180 && resolution <= 360) {
      return 360;
    } else if (resolution > 360 && resolution <= 720) {
      return 720;
    } else if (resolution > 720 && resolution <= 1080) {
      return 1080;
    } else if (resolution > 1080 && resolution <= 1440) {
      return 1440;
    } else {
      return 2160;
    }
  }
}

enum _ProducerType { micProducer, cameraProducer, screenshareProducer }
