import 'dart:developer';
import 'dart:math' as Math;

import 'package:videosdk/src/core/webrtc/src/consumer.dart';

class SdkCapabilities {
  static List<dynamic> getResolutionScalings(List<dynamic> encodings) {
    List<dynamic> resolutionScalings = [];

    if (encodings.length == 0) {
      return resolutionScalings;
    }

    if (encodings.length == 1) {
      const spatialLayers = 3;
      for (int i = 0; i < spatialLayers; i++) {
        resolutionScalings.add(Math.pow(2, (spatialLayers - i - 1)));
      }

      return resolutionScalings;
    }

    // Simulcast encodings
    bool scaleResolutionDownByDefined = false;

    encodings.forEach((encoding) {
      if (encoding['scaleResolutionDownBy'] != null) {
        // at least one scaleResolutionDownBy is defined
        scaleResolutionDownByDefined = true;
        // scaleResolutionDownBy must be >= 1.0
        resolutionScalings
            .add(Math.max(1, encoding['scaleResolutionDownBy'] as num));
      } else {
        // If encodings contains any encoding whose scaleResolutionDownBy
        // attribute is defined, set any undefined scaleResolutionDownBy
        // of the other encodings to 1.0.
        resolutionScalings.add(1.0);
      }
    });

    // If the scaleResolutionDownBy attribues of sendEncodings are
    // still undefined, initialize each encoding's scaleResolutionDownBy
    // to 2^(length of sendEncodings - encoding index - 1).
    if (!scaleResolutionDownByDefined) {
      for (var i = 0; i < encodings.length; i++) {
        resolutionScalings[i] = Math.pow(2, encodings.length - i - 1);
      }
    }

    return resolutionScalings;
  }

  static Map<String, dynamic>? getAdaptivePreferredLayers(
      Consumer consumer, viewportWidth, viewportHeight) {
    log("viewportWidth $viewportWidth");
    log("viewportHeight $viewportHeight");
    Map<String, dynamic> result = new Map();
    result.putIfAbsent('consumerId', () => consumer.id);
    try {
      int width = consumer.appData['width'];
      int height = consumer.appData['height'];
      List<dynamic> encodings = consumer.appData['encodings'];

      List<dynamic> resolutionScalings = getResolutionScalings(encodings);
      log(resolutionScalings.toString());
      const adaptiveScalingFactor = 0.75;

      int newPreferredSpatialLayer = 0;
      for (int i = 0; i < resolutionScalings.length; i++) {
        var levelWidth = adaptiveScalingFactor * width / resolutionScalings[i];
        var levelHeight =
            adaptiveScalingFactor * height / resolutionScalings[i];
        log("levelWidth = $levelWidth");
        log("levelHeight = $levelHeight");
        if (viewportWidth >= levelWidth && viewportHeight >= levelHeight) {
          newPreferredSpatialLayer = i;
        } else {
          break;
        }
      }

      var newPreferredTemporalLayer = consumer.temporalLayer;
      log(consumer.temporalLayer.toString());

      if (newPreferredSpatialLayer == 0 && newPreferredTemporalLayer > 0) {
        var lowestLevelWidth = width / resolutionScalings[0];
        var lowestLevelHeight = height / resolutionScalings[0];

        if (viewportWidth < lowestLevelWidth * 0.5 &&
            viewportHeight < lowestLevelHeight * 0.5) {
          newPreferredTemporalLayer -= 1;
        }
        if (newPreferredTemporalLayer > 0 &&
            viewportWidth < lowestLevelWidth * 0.25 &&
            viewportHeight < lowestLevelHeight * 0.25) {
          newPreferredTemporalLayer -= 1;
        }
      }

      result.addAll({
        'newPreferredSpatialLayer': newPreferredSpatialLayer,
        'newPreferredTemporalLayer': newPreferredTemporalLayer,
      });
    } catch (exception) {
      log("Exception ${exception.toString()}");
      return null;
    }
    return result;
  }
}
